from copy import deepcopy
from datetime import datetime, timezone
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler
from metafold.assets import Asset
from metafold.jobs import Job
from urllib.parse import parse_qs, urlparse
import json
import pytest

default_dt = datetime(2024, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

asset_json = {
    "id": "1",
    "filename": "f763df409e79eb1c.bin",
    "size": 16777216,
    "checksum": "sha256:6310a5951d58eb3e0fdd8c8767c606615552899e65019cb1582508a7c7bfec39",
    "created": "Mon, 01 Jan 2024 00:00:00 GMT",
    "modified": "Mon, 01 Jan 2024 00:00:00 GMT",
}

asset_obj = Asset(
    id="1",
    filename="f763df409e79eb1c.bin",
    size=16777216,
    checksum="sha256:6310a5951d58eb3e0fdd8c8767c606615552899e65019cb1582508a7c7bfec39",
    created=default_dt,
    modified=default_dt,
)

# Default sort order is descending by id
job_list = [
    {
        "id": "3",
        "name": "bar",
        "type": "evaluate_graph",
        "parameters": {
            "graph": None,
        },
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "state": "success",
        "assets": [asset_json],
        "meta": None,
    },
    {
        "id": "2",
        "name": "foo",
        "type": "evaluate_graph",
        "parameters": {
            "graph": None,
        },
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "state": "success",
        "assets": [asset_json],
        "meta": None,
    },
    {
        "id": "1",
        "name": "foo",
        "type": "evaluate_graph",
        "parameters": {
            "graph": None,
        },
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "state": "success",
        "assets": [asset_json],
        "meta": None,
    },
]

new_job = {
    "id": "1",
    "name": "My Job",
    "type": "test_job",
    "parameters": {
        "foo": 1,
        "bar": "a",
        "baz": [2, "b"],
    },
    "created": "Mon, 01 Jan 2024 00:00:00 GMT",
    "assets": [],
    "meta": None,
}

poll_count: int = 0


class MockRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        u = urlparse(self.path)
        params = parse_qs(u.query)
        match u.path:
            case "/projects/1/jobs":
                self.send_response(HTTPStatus.OK)
                self.send_header("Content-Type", "application/json")
                self.end_headers()
                payload = job_list
                if params.get("sort") == ["id:1"]:
                    payload = sorted(job_list, key=lambda p: p["id"])
                elif params.get("q") == ["name:foo"]:
                    payload = [p for p in job_list if p["name"] == "foo"]
                self.wfile.write(json.dumps(payload).encode())
            case "/projects/1/jobs/1":
                self.send_response(HTTPStatus.OK)
                self.send_header("Content-Type", "application/json")
                self.end_headers()
                payload = job_list[-1]
                self.wfile.write(json.dumps(payload).encode())
            case "/projects/1/jobs/1/status":
                global poll_count
                poll_count += 1
                if poll_count < 3:
                    self.send_response(HTTPStatus.ACCEPTED)
                    payload = deepcopy(new_job)
                    payload["state"] = "started"
                else:
                    self.send_response(HTTPStatus.OK)
                    payload = deepcopy(new_job)
                    payload.update({
                        "state": "success",
                        "assets": [asset_json],
                    })
                self.send_header("Content-Type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(payload).encode())
            case _:
                self.send_error(HTTPStatus.NOT_FOUND)

    def do_POST(self):
        match self.path:
            case "/projects/1/jobs":
                self.send_response(HTTPStatus.ACCEPTED)
                self.send_header("Content-Type", "application/json")
                self.end_headers()
                payload = deepcopy(new_job)
                payload.update({
                    "state": "pending",
                    "link": "http://localhost:8000/projects/1/jobs/1/status",
                })
                self.wfile.write(json.dumps(payload).encode())
            case _:
                self.send_error(HTTPStatus.NOT_FOUND)

    def do_PATCH(self):
        match self.path:
            case "/projects/1/jobs/1":
                self.send_response(HTTPStatus.OK)
                self.send_header("Content-Type", "application/json")
                self.end_headers()
                payload = deepcopy(job_list[-1])
                payload["name"] = "baz"
                self.wfile.write(json.dumps(payload).encode())
            case _:
                self.send_error(HTTPStatus.NOT_FOUND)


@pytest.fixture(scope="module")
def request_handler():
    return MockRequestHandler


def test_list_jobs(client):
    jobs = client.jobs.list()
    assert [j.id for j in jobs] == ["3", "2", "1"]


def test_list_jobs_sorted(client):
    jobs = client.jobs.list(sort="id:1")
    assert [j.id for j in jobs] == ["1", "2", "3"]


def test_list_jobs_filtered(client):
    jobs = client.jobs.list(q="name:foo")
    assert all([j.name == "foo" for j in jobs])


def test_get_job(client):
    j = client.jobs.get("1")
    assert j == Job(
        id="1",
        name="foo",
        type="evaluate_graph",
        parameters={
            "graph": None,
        },
        created=default_dt,
        state="success",
        assets=[asset_obj],
        meta=None,
    )


def test_run_job(client):
    params = {
        "foo": 1,
        "bar": "a",
        "baz": [2, "b"],
    }
    j = client.jobs.run("test_job", params, name="My Job")
    assert j == Job(
        id="1",
        name="My Job",
        type="test_job",
        parameters=params,
        created=default_dt,
        state="success",
        assets=[asset_obj],
        meta=None,
    )


def test_update_job(client):
    j = client.jobs.update("1", name="baz")
    assert j.name == "baz"



from attrs import field, frozen
from datetime import datetime
from metafold.api import asdatetime, asdict
from metafold.assets import Asset
from metafold.client import Client
from metafold.exceptions import PollTimeout
from requests import Response
from typing import Any, Optional
import time


def _assets(v: list[dict[str, Any] | Asset]) -> list[Asset]:
    return [a if isinstance(a, Asset) else Asset(**a) for a in v]


@frozen(kw_only=True)
class Job:
    """Job resource.

    Attributes:
        id: Job ID.
        name: Job name.
        type: Job type.
        parameters: Job parameters.
        created: Job creation datetime.
        state: Job state. May be one of: pending, started, success, or failure.
        assets: List of generated asset resources.
        meta: Additional metadata generated by the job.
    """
    id: str
    name: Optional[str] = None
    type: str
    parameters: dict[str, Any]
    created: datetime = field(converter=asdatetime)
    state: str
    assets: list[Asset] = field(converter=_assets)
    meta: dict[str, Any]


class JobsEndpoint:
    """Metafold jobs endpoint."""

    def __init__(self, client: Client) -> None:
        self._client = client

    def list(self, sort: Optional[str] = None, q: Optional[str] = None) -> list[Job]:
        """List jobs.

        Args:
            sort: Sort string. For details on syntax see the Metafold API docs.
                Supported sorting fields are: "id", "name", or "created".
            q: Query string. For details on syntax see the Metafold API docs.
                Supported search fields are: "id", "name", "type", and "state".

        Returns:
            List of job resources.
        """
        url = f"/projects/{self._client.project_id}/jobs"
        payload = asdict(sort=sort, q=q)
        r: Response = self._client.get(url, params=payload)
        return [Job(**j) for j in r.json()]

    def get(self, id: str) -> Job:
        """Get a job.

        Args:
            id: ID of job to get.

        Returns:
            Job resource.
        """
        url = f"/projects/{self._client.project_id}/jobs/{id}"
        r: Response = self._client.get(url)
        return Job(**r.json())

    def run(
        self, type: str, params: dict[str, Any],
        name: Optional[str] = None,
        timeout: int | float = 120,
    ) -> Job:
        """Dispatch a new job and wait for a result.

        See Metafold API docs for the full list of jobs.

        Args:
            type: Job type.
            params: Job parameters.
            name: Optional job name.
            timeout: Time in seconds to wait for a result.

        Returns:
            Completed job resource.
        """
        payload = asdict(type=type, parameters=params, name=name)
        r: Response = self._client.post(f"/projects/{self._client.project_id}/jobs", json=payload)
        r = self._poll(r.json()["link"], timeout)
        return Job(**r.json())

    def update(self, id: str, name: Optional[str] = None) -> Job:
        """Update a job.

        Args:
            id: ID of job to update.
            name: New job name. The existing name remains unchanged if None.

        Returns:
            Updated job resource.
        """
        url = f"/projects/{self._client.project_id}/jobs/{id}"
        payload = asdict(name=name)
        r: Response = self._client.patch(url, data=payload)
        return Job(**r.json())

    def _poll(self, url: str, timeout: int | float) -> Response:
        t0 = time.monotonic()
        r = self._client.get(url)
        while r.status_code == 202:
            elapsed = time.monotonic() - t0
            if elapsed >= timeout:
                raise PollTimeout("Job timed out")
            time.sleep(1)
            r = self._client.get(url)
        return r



from requests import HTTPError, Response, Session
from typing import Any, Callable
from urllib.parse import urljoin
import platform


class Client:
    """Base client."""

    def __init__(self, access_token: str, project_id: str, base_url: str) -> None:
        # TODO(ryan): Validate project id
        self._project_id = project_id
        self._base_url = base_url
        self._session = Session()
        self._session.headers.update({
            "Accept": "application/json",
            "Authorization": f"Bearer {access_token}",
            "User-Agent": f"Python/{platform.python_version()}",
        })

    @property
    def project_id(self) -> str:
        return self._project_id

    def _request(
        self, request: Callable[..., Response], url: str,
        *args: Any, **kwargs: Any,
    ) -> Response:
        url = urljoin(self._base_url, url)
        r: Response = request(url, *args, **kwargs)
        if not r.ok:
            r: dict[str, Any] = r.json()
            # Error responses aren't entirely consistent in the Metafold API,
            # for now we check for a handful of possible fields.
            reason = r.get("errors") or r.get("msg") or r.get("description")
            raise HTTPError(f"HTTP error occurred: {reason or r.reason}")
        return r

    def get(self, url: str, *args: Any, **kwargs: Any) ->  Response:
        return self._request(self._session.get, url, *args, **kwargs)

    def post(self, url: str, *args: Any, **kwargs: Any) ->  Response:
        return self._request(self._session.post, url, *args, **kwargs)

    def put(self, url: str, *args: Any, **kwargs: Any) ->  Response:
        return self._request(self._session.put, url, *args, **kwargs)

    def patch(self, url: str, *args: Any, **kwargs: Any) ->  Response:
        return self._request(self._session.patch, url, *args, **kwargs)

    def delete(self, url: str, *args: Any, **kwargs: Any) ->  Response:
        return self._request(self._session.delete, url, *args, **kwargs)



from datetime import datetime, timezone
from typing import Any


def asdatetime(s: str | datetime) -> datetime:
    """Parse Metafold API datetime.

    Note datetime strings returned by the Metafold API are RFC 1123 formatted,
    times are always in GMT.

    Returns:
        UTC datetime.
    """
    if isinstance(s, datetime):
        return s
    return datetime.strptime(s, "%a, %d %b %Y %H:%M:%S %Z").replace(tzinfo=timezone.utc)


def asdict(**kwargs: Any) -> dict[str, Any]:
    """Convert present kwargs to dictionary."""
    d = {}
    for k, v in kwargs.items():
        if v is not None:
            d[k] = v
    return d



from metafold.client import Client
from metafold.assets import AssetsEndpoint
from metafold.jobs import JobsEndpoint


class MetafoldClient(Client):
    """Metafold REST API client.

    Attributes:
        assets: Sub-client for assets endpoint.
        jobs: Sub-client for jobs endpoint.
    """

    def __init__(
        self, access_token: str, project_id: str,
        base_url: str = "https://api.metafold3d.com",
    ) -> None:
        """Initialize Metafold API client.

        Args:
            access_token: Metafold API secret key.
            project_id: ID of the project to make API calls against.
            base_url: Metafold API URL. Used for internal testing.
        """
        super().__init__(access_token, project_id, base_url=base_url)
        self.assets = AssetsEndpoint(self)
        self.jobs = JobsEndpoint(self)



from copy import deepcopy
from datetime import datetime, timezone
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler
from metafold.assets import Asset
from pathlib import Path
from requests_toolbelt import MultipartDecoder
from urllib.parse import parse_qs, urlparse
import filecmp
import json
import pytest

test_root = Path(__file__).parent
test_file = test_root / "test.png"

default_dt = datetime(2024, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

# Default sort order is descending by id
asset_list = [
    {
        "id": "3",
        "filename": "2c7386e81b6d2ed4.bin",
        "size": 16777216,
        "checksum": "sha256:b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f4850b878ae4944c",
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "modified": "Mon, 01 Jan 2024 00:00:00 GMT",
    },
    {
        "id": "2",
        "filename": "f763df409e79eb1c.bin",
        "size": 16777216,
        "checksum": "sha256:6310a5951d58eb3e0fdd8c8767c606615552899e65019cb1582508a7c7bfec39",
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "modified": "Mon, 01 Jan 2024 00:00:00 GMT",
    },
    {
        "id": "1",
        "filename": "f763df409e79eb1c.bin",
        "size": 16777216,
        "checksum": "sha256:6310a5951d58eb3e0fdd8c8767c606615552899e65019cb1582508a7c7bfec39",
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "modified": "Mon, 01 Jan 2024 00:00:00 GMT",
    },
]

new_asset = {
    "id": "1",
    "filename": "test.png",
    "size": 67,
    "checksum": "sha256:089ad5bf4831b6758e9907db43bc5ebba2e9248a9929dad6132c49932e538278",
    "created": "Mon, 01 Jan 2024 00:00:00 GMT",
    "modified": "Mon, 01 Jan 2024 00:00:00 GMT",
}


class MockRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        u = urlparse(self.path)
        params = parse_qs(u.query)
        match u.path:
            case "/projects/1/assets":
                self.send_response(HTTPStatus.OK)
                self.send_header("Content-Type", "application/json")
                self.end_headers()
                payload = asset_list
                if params.get("sort") == ["id:1"]:
                    payload = sorted(asset_list, key=lambda p: p["id"])
                elif params.get("q") == ["filename:f763df409e79eb1c.bin"]:
                    payload = [p for p in asset_list if p["filename"] == "f763df409e79eb1c.bin"]
                self.wfile.write(json.dumps(payload).encode())
            case "/projects/1/assets/1":
                self.send_response(HTTPStatus.OK)
                self.send_header("Content-Type", "application/json")
                self.end_headers()
                payload = deepcopy(asset_list[-1])
                if params.get("download") == ["true"]:
                    payload["link"] = "http://localhost:8000/download?filename=f763df409e79eb1c.bin"
                self.wfile.write(json.dumps(payload).encode())
            case "/download":
                self.send_response(HTTPStatus.OK)
                self.end_headers()
                with open(test_file, "rb") as f:
                    self.wfile.write(f.read())
            case _:
                self.send_error(HTTPStatus.NOT_FOUND)

    def do_POST(self):
        match self.path:
            case "/projects/1/assets":
                self.send_response(HTTPStatus.ACCEPTED)
                self.send_header("Content-Type", "application/json")
                self.end_headers()
                self._assert_file()
                self.wfile.write(json.dumps(new_asset).encode())
            case _:
                self.send_error(HTTPStatus.NOT_FOUND)

    def do_PATCH(self):
        match self.path:
            case "/projects/1/assets/1":
                self.send_response(HTTPStatus.OK)
                self.send_header("Content-Type", "application/json")
                self.end_headers()
                self._assert_file()
                self.wfile.write(json.dumps(new_asset).encode())
            case _:
                self.send_error(HTTPStatus.NOT_FOUND)

    def do_DELETE(self):
        match self.path:
            case "/projects/1/assets/1":
                self.send_response(HTTPStatus.OK)
                self.end_headers()
            case _:
                self.send_error(HTTPStatus.NOT_FOUND)

    def _assert_file(self):
        length = self.headers.get("Content-Length")
        content_type = self.headers.get("Content-Type")
        content = self.rfile.read(int(length))
        decoder = MultipartDecoder(content, content_type)
        file = decoder.parts[0].content
        with open(test_file, "rb") as f:
            assert file == f.read()


@pytest.fixture(scope="module")
def request_handler():
    return MockRequestHandler


def test_list_assets(client):
    assets = client.assets.list()
    assert [a.id for a in assets] == ["3", "2", "1"]


def test_list_assets_sorted(client):
    assets = client.assets.list(sort="id:1")
    assert [a.id for a in assets] == ["1", "2", "3"]


def test_list_assets_filtered(client):
    assets = client.assets.list(q="filename:f763df409e79eb1c.bin")
    assert all([a.filename == "f763df409e79eb1c.bin" for a in assets])


def test_get_asset(client):
    a = client.assets.get("1")
    assert a == Asset(
        id="1",
        filename="f763df409e79eb1c.bin",
        size=16777216,
        checksum="sha256:6310a5951d58eb3e0fdd8c8767c606615552899e65019cb1582508a7c7bfec39",
        created=default_dt,
        modified=default_dt,
    )


def test_download_asset(client, tmp_path):
    path = tmp_path / "test.png"
    client.assets.download_file("1", path)
    assert filecmp.cmp(path, test_file, shallow=False)


def test_create_asset(client):
    with open(test_file, "rb") as f:
        a = client.assets.create(f)
    assert a == Asset(
        id="1",
        filename="test.png",
        size=67,
        checksum="sha256:089ad5bf4831b6758e9907db43bc5ebba2e9248a9929dad6132c49932e538278",
        created=default_dt,
        modified=default_dt,
    )


def test_update_asset(client):
    with open(test_file, "rb") as f:
        a = client.assets.update("1", f)
    assert a == Asset(
        id="1",
        filename="test.png",
        size=67,
        checksum="sha256:089ad5bf4831b6758e9907db43bc5ebba2e9248a9929dad6132c49932e538278",
        created=default_dt,
        modified=default_dt,
    )


def test_delete_asset(client):
    # FIXME: Assert something
    client.assets.delete("1")



from attrs import field, frozen
from datetime import datetime
from metafold.api import asdatetime, asdict
from metafold.client import Client
from os import PathLike
from requests import Response
from typing import IO, Optional
import requests


@frozen(kw_only=True)
class Asset:
    """Asset resource.

    Attributes:
        id: Asset ID.
        filename: Asset filename.
        size: File size in bytes.
        checksum: File checksum.
        created: Asset creation datetime.
        modified: Asset last modified datetime.
    """
    id: str
    filename: str
    size: int
    checksum: str
    created: datetime = field(converter=asdatetime)
    modified: datetime = field(converter=asdatetime)


class AssetsEndpoint:
    """Metafold assets endpoint."""

    def __init__(self, client: Client) -> None:
        self._client = client

    def list(self, sort: Optional[str] = None, q: Optional[str] = None) -> list[Asset]:
        """List assets.

        Args:
            sort: Sort string. For details on syntax see the Metafold API docs.
                Supported sorting fields are: "id", "filename", "size", "created", or
                "modified".
            q: Query string. For details on syntax see the Metafold API docs.
                Supported search fields are: "id" and "filename".

        Returns:
            List of asset resources.
        """
        url = f"/projects/{self._client.project_id}/assets"
        payload = asdict(sort=sort, q=q)
        r: Response = self._client.get(url, params=payload)
        return [Asset(**a) for a in r.json()]

    def get(self, id: str) -> Asset:
        """Get an asset.

        Args:
            id: ID of asset to get.

        Returns:
            Asset resource.
        """
        url = f"/projects/{self._client.project_id}/assets/{id}"
        r: Response = self._client.get(url)
        return Asset(**r.json())

    def download_file(self, id: str, path: str | PathLike):
        """Download an asset.

        Args:
            id: ID of asset to download.
            path: Path to downloaded file.
        """
        url = f"/projects/{self._client.project_id}/assets/{id}"
        r: Response = self._client.get(url, params={"download": "true"})
        r = requests.get(r.json()["link"], stream=True)
        with open(path, "wb") as f:
            for chunk in r.iter_content(chunk_size=65536):  # 64 KiB
                f.write(chunk)

    def create(self, f: str | bytes | PathLike | IO[bytes]) -> Asset:
        """Upload an asset.

        Args:
            f: File-like object (opened in binary mode) or path to file on disk.

        Returns:
            Asset resource.
        """
        fp: IO[bytes] = _open_file(f)
        try:
            url = f"/projects/{self._client.project_id}/assets"
            r: Response = self._client.post(url, files={"file": fp})
        finally:
            fp.close()
        return Asset(**r.json())

    def update(self, id: str, f: str | bytes | PathLike | IO[bytes]) -> Asset:
        """Update an asset.

        Args:
            id: ID of asset to update.
            f: File-like object (opened in binary mode) or path to file on disk.

        Returns:
            Updated asset resource.
        """
        fp: IO[bytes] = _open_file(f)
        try:
            url = f"/projects/{self._client.project_id}/assets/{id}"
            r: Response = self._client.patch(url, files={"file": fp})
        finally:
            fp.close()
        return Asset(**r.json())

    def delete(self, id: str) -> None:
        """Delete an asset.

        Args:
            id: ID of asset to delete.
        """
        url = f"/projects/{self._client.project_id}/assets/{id}"
        self._client.delete(url)


def _open_file(f: str | bytes | PathLike | IO[bytes]) -> IO[bytes]:
    if isinstance(f, str | bytes | PathLike):
        return open(f, "rb")
    return f
